/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package fr.amap.lidar.amapvox;

import fr.amap.lidar.amapvox.gui.JavafxApp;
import fr.amap.lidar.amapvox.commons.Configuration;
import fr.amap.lidar.amapvox.gui.Global;
import fr.amap.lidar.amapvox.gui.Util;
import fr.amap.lidar.amapvox.simulation.hemi.HemiScanView;
import fr.amap.lidar.amapvox.simulation.transmittance.TransmittanceSim;
import fr.amap.lidar.amapvox.simulation.transmittance.lai2xxx.Lai2xxxSim;
import fr.amap.lidar.amapvox.util.AVoxTask;
import fr.amap.lidar.amapvox.voxelisation.als.LasVoxelisation;
import fr.amap.lidar.amapvox.voxelisation.configuration.TLSVoxCfg;
import fr.amap.lidar.amapvox.voxelisation.postproc.ButterflyRemover;
import fr.amap.lidar.amapvox.voxelisation.postproc.VoxelFileMerging;
import fr.amap.lidar.amapvox.voxelisation.tls.PTGVoxelisation;
import fr.amap.lidar.amapvox.voxelisation.tls.PTXVoxelisation;
import fr.amap.lidar.amapvox.voxelisation.tls.RSPVoxelisation;
import fr.amap.lidar.amapvox.voxelisation.tls.RxpVoxelisation;
import java.io.File;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javafx.application.Application;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * AMAPVox main class that either launches the JavaFX application or run the
 * simulations from command line arguments.
 *
 * @author Philippe Verley (philippe.verley@ird.fr)
 */
public class Launcher {

    private final static Logger LOGGER = LoggerFactory.getLogger(Launcher.class);

    // valid command line arguments
    private final static String[] ARGS = new String[]{
        "--version", "-h", "--execute-cfg=", "--T=", "--T-TLS_VOX="
    };

    public static void main(String[] args) throws Exception {

        boolean nogui = false;
        for (String arg : args) {
            if (arg.equals("-h")) {
                showHelp();
                System.exit(0);
            }
            if (arg.equals("--version")) {
                System.out.println("amapvox version " + Global.buildVersion);
                System.exit(0);
            }
            if (arg.startsWith("--execute-cfg")) {
                nogui = true;
            }
        }

        if (nogui) {
            // check for unknown arguments
            List<String> invalidArgs = new ArrayList();
            for (String arg : args) {
                boolean ok = false;
                for (String validArg : ARGS) {
                    ok |= arg.startsWith(validArg);
                }
                if (!ok) {
                    invalidArgs.add(arg);
                }
            }
            if (!invalidArgs.isEmpty()) {
                LOGGER.error("Invalid command line arguments: " + Arrays.toString(invalidArgs.toArray()));
                showHelp();
                System.exit(0);
            }

            // run AMAPVox without GUI
            new Launcher().runWithoutGUI(args);
        } else {
            if (args.length > 0) {
                LOGGER.warn("Command line arguments are ignored in GUI mode. Ignored arguments: " + Arrays.toString(args));
            }
            // start JavaFX application
            Application.launch(JavafxApp.class, args);
        }
    }

    /**
     * Run AMAPVox without GUI from command line arguments.
     *
     * @param args, a list of command line arguments
     * @throws Exception
     */
    private void runWithoutGUI(String[] args) throws Exception {

        String[] cfgFileArray = new String[0];
        int nCpuTlsVox = 1;
        int nCpu = 1;

        for (String arg : args) {
            // all argument but -h (handled in main function must comply with format name=value(s)
            if (arg.split("=").length != 2) {
                System.err.println("Wrong command line format");
                showHelp();
            }
            // --execute-cfg
            if (arg.startsWith("--execute-cfg=")) {
                String[] executeCfgArg = arg.split("=");
                cfgFileArray = executeCfgArg[1].split(" ");
            }
            // --T
            if (arg.startsWith("--T=")) {
                nCpu = Math.min(getNCpuFromArg(arg), cfgFileArray.length);
            }

            // --T-TLS_VOX
            if (arg.startsWith("--T-TLS_VOX=")) {
                nCpuTlsVox = getNCpuFromArg(arg);
            }
        }

        // create and init tasks
        List<AVoxTask> taskList = new ArrayList();
        for (String filePath : cfgFileArray) {
            taskList.add(provideTask(new File(filePath), nCpuTlsVox));
        }
        for (AVoxTask task : taskList) {
            task.init();
        }

        StringBuilder sb = new StringBuilder();
        sb.append("Execute task(s) ");
        if (nCpu > 1) {
            sb.append("concurrently (nThread=");
            sb.append(nCpu);
            sb.append(")");
        } else {
            sb.append("sequentially");
        }
        LOGGER.info(sb.toString());

        // run task list
        ExecutorService exec = Executors.newFixedThreadPool(nCpu);
        List<Future<File[]>> results = exec.invokeAll(taskList);

        // output summary
        LOGGER.info("OUTPUT SUMMARY");
        SimpleDateFormat sdf = new SimpleDateFormat("yyyy MMM dd HH:mm");
        for (int itask = 0; itask < taskList.size(); itask++) {
            LOGGER.info(taskList.get(itask).getName() + " " + taskList.get(itask).getFile().getName());
            for (File file : results.get(itask).get()) {
                sb = new StringBuilder();
                sb.append("  (");
                sb.append(sdf.format(file.lastModified()));
                sb.append(", ");
                sb.append(Util.humanReadableByteCount(file.length(), true));
                sb.append(") ");
                sb.append(file.getAbsolutePath());
                LOGGER.info(sb.toString());
            }
        }

        // exit
        LOGGER.info("Simulation(s) completed successfully");
        System.exit(0);
    }

    /**
     * Returns a {@link fr.amap.lidar.amapvox.util.AVoxTask} associated to the
     * given configuration file and allocates {@code ncpu} threads to the task.
     *
     * @param file, the XML configuration file.
     * @param ncpu, the number of threads allocated to the task
     * @return the {@link fr.amap.lidar.amapvox.util.AVoxTask} associated to the
     * configuration file.
     * @throws Exception
     */
    private AVoxTask provideTask(File file, int ncpu) throws Exception {

        String type = Configuration.readType(file);

        switch (type) {
            case "voxelisation-ALS":
                return createTask(LasVoxelisation.class, file, 1);
            case "voxelisation-TLS":
                final TLSVoxCfg tLSVoxCfg = new TLSVoxCfg();
                try {
                    tLSVoxCfg.readConfiguration(file);
                } catch (Exception ex) {
                    LOGGER.error("Failed to read configuration file " + file.getAbsolutePath(), ex);
                    return null;
                }
                switch (tLSVoxCfg.getInputType()) {
                    case RSP_PROJECT:
                        return createTask(RSPVoxelisation.class, file, ncpu);
                    case RXP_SCAN:
                        return createTask(RxpVoxelisation.class, file, ncpu);
                    case PTG_PROJECT:
                        return createTask(PTGVoxelisation.class, file, ncpu);
                    case PTX_PROJECT:
                        return createTask(PTXVoxelisation.class, file, ncpu);
                }
            case "transmittance":
                return createTask(TransmittanceSim.class, file, 1);
            case "LAI2000":
            case "LAI2200":
                return createTask(Lai2xxxSim.class, file, 1);
            case "Hemi-Photo":
                return createTask(HemiScanView.class, file, 1);
            case "merging":
                return createTask(VoxelFileMerging.class, file, 1);
            case "butterfly-removing":
                return createTask(ButterflyRemover.class, file, 1);
        }

        LOGGER.error("Undefined type of configuration file: " + type);
        return null;
    }

    /**
     * Creates a custom {@link fr.amap.lidar.amapvox.util.AVoxTask} based on the
     * class name, the configuration file and a number of allocated threads.
     *
     * @param className
     * @param file
     * @param ncpu
     * @return
     * @throws Exception
     */
    private AVoxTask createTask(Class className, File file, int ncpu) throws Exception {

        return (AVoxTask) className.getConstructor(File.class, int.class).newInstance(file, ncpu);
    }

    /**
     * Extract the number of threads from the command line argument. Expected
     * argument format --name=integer. For number of threads set to zero the
     * function returns maximum number of available processors.
     *
     * @param arg
     * @return
     */
    private static int getNCpuFromArg(String arg) {
        try {
            String[] nCpuArg = arg.split("=");
            return nCpuArg[1].equals("0")
                    ? Runtime.getRuntime().availableProcessors()
                    : Integer.valueOf(nCpuArg[1]);
        } catch (Exception ex) {
            LOGGER.error("Wrong format for argument " + arg);
            showHelp();
            System.exit(1);
        }
        return -1;
    }

    /**
     * Print help message in standard output.
     */
    public static void showHelp() {

        StringBuilder sb = new StringBuilder();
        sb.append("Valid command line options:\n");
        sb.append("  -h            ");
        sb.append("print this help message.\n");
        sb.append("  --version     ");
        sb.append("print product version and exit.\n");
        sb.append("  --execute-cfg ");
        sb.append("run task list from commande line.\n");
        sb.append("                ");
        sb.append("  --execute-cfg=\"/home/file1.xml /home/file2.xml /home/file3.xml /home/file4.xml\"\n");
        sb.append("  --T           ");
        sb.append("number of threads allocated to task execution. By default T=1.\n");
        sb.append("                ");
        sb.append("  --T=4 run task list with 4 threads.\n");
        sb.append("                ");
        sb.append("  --T=0 run task list with as many threads as possible.\n");
        sb.append("  --T-TLS_VOX   ");
        sb.append("number of threads allocated to TLS voxelisation task. By default T-TLS_VOX=1\n");
        sb.append("                ");
        sb.append("  --T-TLS_VOX=8 run TLS voxelisation with 8 threads.\n");
        sb.append("                ");
        sb.append("  --T-TLS_VOX=0 run voxelisation with as many threads as possible.\n");

        System.out.println(sb.toString());
    }

}
